﻿namespace TabsStudioSync
{
    class SolutionExplorer
    {
        public SolutionExplorer(EnvDTE80.DTE2 dte_)
        {
            dte = dte_;
        }

        public void SyncDocument(EnvDTE.Document doc)
        {
            ActivateSolutionExplorer();

            EnvDTE.ProjectItem projectItem = doc.ProjectItem;
            if (projectItem == null)
                return;

            EnvDTE.UIHierarchyItem target = Expand(projectItem);
            target.Select(EnvDTE.vsUISelectionType.vsUISelectionTypeSelect);
        }

        public void SyncDocuments(System.Collections.Generic.List<EnvDTE.Document> docs)
        {
            ActivateSolutionExplorer();

            bool firstItem = true;
            foreach (EnvDTE.Document doc in docs)
            {
                EnvDTE.ProjectItem projectItem = doc.ProjectItem;
                if (projectItem == null)
                    continue;

                EnvDTE.UIHierarchyItem target = Expand(projectItem);
                if (firstItem)
                {
                    target.Select(EnvDTE.vsUISelectionType.vsUISelectionTypeSelect);
                    firstItem = false;
                }
                else
                {
                    if (!target.IsSelected)
                        target.Select(EnvDTE.vsUISelectionType.vsUISelectionTypeToggle);
                }
            }
        }

        private EnvDTE.UIHierarchyItem Expand(EnvDTE.ProjectItem projectItem)
        {
            return ExpandPath(dte.ToolWindows.SolutionExplorer.UIHierarchyItems.Item(1).UIHierarchyItems, GetPath(projectItem));
        }

        private void ActivateSolutionExplorer()
        {
            dte.ToolWindows.SolutionExplorer.Parent.Activate();
        }

        private System.Collections.Generic.List<object> GetPath(object p)
        {
            if (p == null)
                return null;

            System.Collections.Generic.List<object> result = new System.Collections.Generic.List<object>();
            if (p is EnvDTE.ProjectItem)
            {
                EnvDTE.ProjectItem projectItem = p as EnvDTE.ProjectItem;
                if (projectItem.Collection != null)
                {
                    System.Collections.Generic.List<object> parents = GetPath(projectItem.Collection.Parent);
                    if (parents != null)
                        result = parents;
                }
            }
            if (p is EnvDTE.Project)
            {
                EnvDTE.Project project = p as EnvDTE.Project;

                object parent = null;
                if (project.ParentProjectItem != null)
                {
                    parent = project.ParentProjectItem;
                }
                else
                {
                    if (!IsRootProject(project))
                        parent = FindParentProject(project);
                }

                System.Collections.Generic.List<object> parents = GetPath(parent);
                if (parents != null)
                    result = parents;
            }
            result.Add(p);
            return result;
        }

        private EnvDTE.UIHierarchyItem ExpandPath(EnvDTE.UIHierarchyItems root, System.Collections.Generic.List<object> path)
        {
            EnvDTE.UIHierarchyItem result = null;
            EnvDTE.UIHierarchyItems currentRoot = root;
            for (int i = 0; i < path.Count; ++i)
            {
                EnvDTE.UIHierarchyItem h = FindObject(currentRoot, path[i]);
                if (h == null)
                    h = FindName(currentRoot, path[i]);
                if (h == null)
                    continue;

                if (i == path.Count - 1)
                {
                    result = h;
                }
                else
                {
                    h.UIHierarchyItems.Expanded = true;
                    currentRoot = h.UIHierarchyItems;
                }
            }
            return result;
        }

        private static EnvDTE.UIHierarchyItem FindObject(EnvDTE.UIHierarchyItems root, object o)
        {
            foreach (EnvDTE.UIHierarchyItem h in root)
            {
                if (h.Object == o)
                    return h;
            }
            return null;
        }

        private static EnvDTE.UIHierarchyItem FindName(EnvDTE.UIHierarchyItems root, object o)
        {
            var p = o as EnvDTE.ProjectItem;
            if (p == null)
                return null;

            foreach (EnvDTE.UIHierarchyItem h in root)
            {
                if (h.Name == p.Name)
                    return h;
            }
            return null;
        }

        private bool IsRootProject(EnvDTE.Project child)
        {
            if (dte.Solution == null || dte.Solution.Projects == null)
                return true;

            foreach (EnvDTE.Project root in dte.Solution.Projects)
            {
                if (root == child)
                    return true;
            }
            return false;
        }

        private EnvDTE.Project FindParentProject(EnvDTE.Project child)
        {
            if (dte.Solution == null || dte.Solution.Projects == null)
                return null;

            foreach (EnvDTE.Project root in dte.Solution.Projects)
            {
                EnvDTE.Project result = FindParentProjectInHierarchy(root, child);
                if (result != null)
                    return result;
            }
            return null;
        }

        private EnvDTE.Project FindParentProjectInHierarchy(EnvDTE.Project root, EnvDTE.Project child)
        {
            if (root != null && root.Kind == VsProjectItemKindSolutionFolder && root.ProjectItems != null)
            {
                foreach (EnvDTE.ProjectItem item in root.ProjectItems)
                {
                    if (item != null && item.Object is EnvDTE.Project)
                    {
                        EnvDTE.Project project = item.Object as EnvDTE.Project;
                        if (project == child)
                            return root;
                        EnvDTE.Project result = FindParentProjectInHierarchy(project, child);
                        if (result != null)
                            return result;
                    }
                }
            }
            return null;
        }

        private const string VsProjectItemKindSolutionFolder = "{66A26720-8FB5-11D2-AA7E-00C04F688DDE}";
        private EnvDTE80.DTE2 dte;
    }
}
